#include "Graphics.h"
#include <iostream.h>
#include <math.h>

/*
 *	The drawing routing for GLUT.  
 *	Draw the messages and call for the splines in this routine.
 */
void Graphics::Display()
{
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	// set the camera position, where we're looking at, and the up vector
	gluLookAt(cameraPos[0], cameraPos[1], cameraPos[2],  0, 0, 0,  0, 1, 0);	
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	if(spline[currentLevel]->renderMode == GL_SELECT)	// we want to pick a control point
	{
		glGetIntegerv(GL_VIEWPORT, viewport);		// get the viewport
		glSelectBuffer(BUFSIZE, selectBuf);		
		(void) glRenderMode(GL_SELECT);			// switch rendering modes
		glInitNames();					// initialize the name stack
	// we need a number on the stack already, so we don't get an error later with our first glLoadName()
		glPushName(0);
	}
	
	glMatrixMode(GL_PROJECTION); 
	glLoadIdentity();
	
	// If we are in selection mode, set the picking matrix
	if(spline[currentLevel]->renderMode == GL_SELECT)
		// specifies a unit cube to search for a hit with the mouse
		gluPickMatrix((GLfloat)mousePos[0], (GLfloat)(viewport[3]-mousePos[1]), 8, 8, viewport);

	float left = -5, bottom=-5;	// the left and bottom of the window
	glOrtho(left, -left, bottom, -bottom, cameraPos[2], 
			currentLevel==0? cameraPos[2]-1 : cameraPos[2]-currentLevel);
	
	glMatrixMode(GL_MODELVIEW);
	
	// draw the current level and the 3 beneath it
	for(int i = currentLevel-3; i<=currentLevel; i++)
	{
		if(i<0)	// the current level is beneath 0 (shouldn't ever happen actually)
			continue;
		if(i==currentLevel-3)		// the bottommost layer we see is blue
			glColor3f(0, 0, 1);
		else if(i==currentLevel-2)	// the second layer from the bottom is purple
			glColor3f(.8, 0, .8);	
		else if(i==currentLevel-1)	// the second layer from the top is teal
			glColor3f(.2, .6, .9);
		else				// the top layer is red
			glColor3f(.9, 0, .1);

		spline[i]->drawSpline(currentLevel);	// draw the current level's spline
	}
	// if we are in selection mode, we need to process the picking buffer
	if(spline[currentLevel]->renderMode==GL_SELECT)
	{
		hits=glRenderMode(GL_RENDER);		// get the hits
		ProcessHits(hits, selectBuf);		// deal with the hits
	}
	
	if(showMessages)	// show the writing on the wall
		Messages();
	
	glutSwapBuffers();
}

/*
 *	Called by Glut when the window is resized.
 *	@param width The width of the window
 *	@param height The height of the window
 */
void Graphics::Reshape(int width, int height)
{
	// Set these for later use in DragTrack()
	WindowWidth = width; WindowHeight = height;
	// Set the viewport
	glViewport(0, 0, (GLsizei)width, (GLsizei)height);
	// Each level of splines needs to know the window size
	for(int x = 0; x < spline.size(); x++)
		spline[x]->setWindowSize(width, height);
}

/*
 *	Called by Glut when the user presses one of the mouse keys
 *	@param button The button that was used
 *	@param state The state of the button
 *	@param x The x coordinate of the mouse
 *	@param y The y coordinate of the mouse
 */
void Graphics::Mouse(int button, int state, int x, int y)
{
	// the user pressed the right mouse button, so lay down a new control point
	if(button == GLUT_RIGHT_BUTTON && state == GLUT_DOWN)	
		spline[currentLevel]->addControlPoint(x, y);

	// we want to pick a point
	if(button == GLUT_LEFT_BUTTON && state == GLUT_DOWN)
	{
		spline[currentLevel]->renderMode = GL_SELECT;	// we're switching to selection mode
		movingMode=1;	// we can drag around the control point we've selected
		Display();	// draw everything in selection mode
		spline[currentLevel]->renderMode = GL_RENDER;	// go back to render mode
	}

	// the user let up the left mouse button, so get out of picking mode
	if(button == GLUT_LEFT_BUTTON && state == GLUT_UP)
	{
		movingMode=0;
		spline[currentLevel]->objectPicked = -1;	// the object is no longer movable
	}

	// we want to delete a node
	if(button == GLUT_MIDDLE_BUTTON && state == GLUT_DOWN)
	{
		deleteMode=1;
		spline[currentLevel]->renderMode = GL_SELECT;
		Display();	// render the scene in selection mode
		spline[currentLevel]->renderMode = GL_RENDER;
	}
}

/*
 *	When Glut has nothing better to do, update the control point position
 *	if we are moving it around and redraw everything.
 */
void Graphics::Idle()
{
	if(movingMode)
		DragTrack();
	glutPostRedisplay();
}

/*	Determine which object was chosen while in picking mode
 *	@param hits	The number of hits we had
 *	@param buffer The selection buffer
 */
void Graphics::ProcessHits(GLint hits, GLuint buffer[])
{
	int pick;
	for(int x = 1; x <= hits; x++)
	{
		pick = buffer[x*4 - 1];	// every fourth item in the buffer is an object number that was hit
		if(pick>0)	// if pick is less than or equal to 0, we haven't selected an object
		{
			spline[currentLevel]->objectPicked=pick-1;
			break;
		}
	}
	if (pick <= 0) spline[currentLevel]->objectPicked=-1;
	else if (deleteMode)
	{
		spline[currentLevel]->deleteControlPoint();
		deleteMode=0;
	}
	
	// We want info on the node that we are interacting with
	else if (movingMode)
	{
		int curve;	// the curve the node is in
		// go through every curve on this level until we reach the one
		// that contains the node that we selected
		for(int i = 0; i < spline[currentLevel]->levelCurves.size(); i++)
		{
			// the node we picked is in level i
			if (spline[currentLevel]->levelCurves[i].controlPoints.size() > pick-1)
			{
				curve = i;
				break;
			}
			
			// the number of nodes in curve i is less than the number of the
			// object we selected, so subtract the number of control points
			// in this curve and move on to the next curve
			else pick -= spline[currentLevel]->levelCurves[i].controlPoints.size();
		}
		printf("Curve number: %d\t\tPoint number: %d\n", i, pick-1);
	}
}

/*
 *	Move the control point around if we are in moving mode
 */
void Graphics::DragTrack()
{
	if(spline[currentLevel]->objectPicked !=-1)
	{
		int objectNumber = spline[currentLevel]->objectPicked;
		// convert the mouse position from pixel locations into real coordinates
		float xcoord = (mousePos[0] / (float)WindowWidth) * 10. - 5.;
		float ycoord = 5. - (mousePos[1] / (float)WindowHeight) * 10.;
		
		int curve;
		// same conversion as that in ProcessHits
		for(int i = 0; i < spline[currentLevel]->levelCurves.size(); i++)
		{
			if (spline[currentLevel]->levelCurves[i].controlPoints.size() > objectNumber)
			{
				curve = i;
				break;
			}
			else objectNumber -= spline[currentLevel]->levelCurves[i].controlPoints.size();
		}
		// store the mouse position into the vertex's coordinates
		spline[currentLevel]->levelCurves[curve].controlPoints[objectNumber].pos[0] = xcoord;
		spline[currentLevel]->levelCurves[curve].controlPoints[objectNumber].pos[1] = ycoord;
	}
}

/*
 *	Write the messages to the screen
 */
void Graphics::Char2Wall(float x,float y,float z, char buf[])
{
	char *p; 
	glRasterPos3f(x,y,z);
    	for(p = buf;*p;p++) 
		glutBitmapCharacter(GLUT_BITMAP_9_BY_15,*p);
}

/*
 *	The messages that we see on the screen.
 */
void Graphics::Messages(void)
{     
	char buf[256]; 
#define  LABEL2(x,y,W,u) {sprintf(buf,(W),(u));Char2Wall(x,y,0.,buf);}
	glMatrixMode(GL_PROJECTION); 
	glPushMatrix(); 
	glLoadIdentity();
	gluOrtho2D(0,3000,0,3000);
	glMatrixMode(GL_MODELVIEW); 
	glPushMatrix(); 
	glLoadIdentity();
    	/* writings */
	glColor3f(.9, 0, .9);
	LABEL2(10,10,"illiLevelEditor by Nachand, U Illinois, 2001 %s","");
	LABEL2(40, 2840, "(F1) to save","");
	LABEL2(600, 2840, "(F12) to open", "");
	glColor3f(1, 0, 0);
    	LABEL2(40,2770,"Level   %i", currentLevel);
	
	// now we are going to match the message color with the color of the level curves
	if(currentLevel>=3)		// we are able to show 4 levels at once
	{
		glColor3f(.2, .6, .9);
		LABEL2(40,2700,"Level   %i", currentLevel-1);
		glColor3f(.8, 0, .8);
		LABEL2(40,2630,"Level   %i", currentLevel-2);
		glColor3f(0, 0, 1);
		LABEL2(40,2560,"Level   %i", currentLevel-3);
	}
	else if(currentLevel==2)	// we can only show three levels
	{
		glColor3f(.2, .6, .9);
		LABEL2(40,2700,"Level   %i", currentLevel-1);
		glColor3f(.8, 0, .8);
		LABEL2(40,2630,"Level   %i", currentLevel-2);
	}
	else if(currentLevel==1)	// we can only show two levels
	{
		glColor3f(.2, .6, .9);
		LABEL2(40,2700,"Level   %i", currentLevel-1);
	}
	glColor3f(.9,0,.9);
	LABEL2(40, 2490, "New (C)urve", "");
	LABEL2(40, 2420, "Cl(O)se curve", "");
    	
	glPopMatrix();
	glMatrixMode(GL_PROJECTION); 
	glPopMatrix();
}

/*
 *	Deals with all ascii keypresses
 */
void Graphics::Keyboard(unsigned char key, int x, int y)
{

#define  IF(K)            if(key==K)
#define  PRESS(K,A,b)     IF(K){b;}IF((K-32)){A;}  
#define  TOGGLE(K,flg)    IF(K){(flg) = 1-(flg); }

	IF(27) { exit(0); }		// ESC exit
	TOGGLE('w',showMessages);	// writing on/off

	if(key=='c' || key=='C')	// add a new curve to the current level
		spline[currentLevel]->addNewCurve();
	if(key=='O')			// close the currently selected curve
		spline[currentLevel]->closeCurve();
	if(key=='o')			// open the currently selected curve
		spline[currentLevel]->openCurve();
	glutPostRedisplay();
}

/*
 *	Handles non-ascii keypresses.
 */
void Graphics::SpecialKeyboard(int key, int x, int y)
{
	switch(key)
	{
		case GLUT_KEY_F1 :	// save to file
		{
			saveData->writeFile(spline);
			break;
		}
		case GLUT_KEY_F12 :	// read from file
		{
			spline.empty();
			saveData->readFile(spline);
			Reshape(glutGet(GLUT_WINDOW_WIDTH), glutGet(GLUT_WINDOW_HEIGHT));
			break;
		}
		case GLUT_KEY_LEFT :	// change focus to the previous curve
		{
			spline[currentLevel]->prevCurve();
			break;
		}
		case GLUT_KEY_RIGHT :	// change focus to the next curve
		{
			spline[currentLevel]->nextCurve();
			break;
		}
		case GLUT_KEY_UP :	// go up one level
		{
			if(spline.size() == (currentLevel+1))
				// we've gone to a level we haven't been before, so copy the last one to it
				spline.push_back(new Splines(*spline[currentLevel]));		
		
			currentLevel++;
			cameraPos[2]++;
			break;
		}
		case GLUT_KEY_DOWN :	// go down one level
		{
			if(currentLevel>0)	
			{
				currentLevel--;
				cameraPos[2]--;
			}
			break;
		}
	}
}
